@@ -4,6 +4,7 @@ import (
 	"fmt"
 	"testing"
 
+	qt "github.com/frankban/quicktest"
 	"github.com/google/go-cmp/cmp"
 )
 
@@ -34,10 +35,10 @@ func TestSizeof(t *testing.T) {
 }
 
 func TestCopyType(t *testing.T) {
-	_ = copyType((*Void)(nil))
+	_, _ = copyType((*Void)(nil), nil)
 
 	in := &Int{Size: 4}
-	out := copyType(in)
+	out, _ := copyType(in, nil)
 
 	in.Size = 8
 	if size := out.(*Int).Size; size != 4 {
@@ -45,7 +46,21 @@ func TestCopyType(t *testing.T) {
 	}
 
 	t.Run("cyclical", func(t *testing.T) {
-		_ = copyType(newCyclicalType(2))
+		_, _ = copyType(newCyclicalType(2), nil)
+	})
+
+	t.Run("identity", func(t *testing.T) {
+		u16 := &Int{Size: 2}
+
+		out, _ := copyType(&Struct{
+			Members: []Member{
+				{Name: "a", Type: u16},
+				{Name: "b", Type: u16},
+			},
+		}, nil)
+
+		outStruct := out.(*Struct)
+		qt.Assert(t, outStruct.Members[0].Type, qt.Equals, outStruct.Members[1].Type)
 	})
 }
 
@@ -54,30 +69,28 @@ func TestCopyType(t *testing.T) {
 // There currently is no better way to document which
 // types implement an interface.
 func ExampleType_validTypes() {
-	var t Type
-	t = &Void{}
-	t = &Int{}
-	t = &Pointer{}
-	t = &Array{}
-	t = &Struct{}
-	t = &Union{}
-	t = &Enum{}
-	t = &Fwd{}
-	t = &Typedef{}
-	t = &Volatile{}
-	t = &Const{}
-	t = &Restrict{}
-	t = &Func{}
-	t = &FuncProto{}
-	t = &Var{}
-	t = &Datasec{}
-	_ = t
+	var _ Type = &Void{}
+	var _ Type = &Int{}
+	var _ Type = &Pointer{}
+	var _ Type = &Array{}
+	var _ Type = &Struct{}
+	var _ Type = &Union{}
+	var _ Type = &Enum{}
+	var _ Type = &Fwd{}
+	var _ Type = &Typedef{}
+	var _ Type = &Volatile{}
+	var _ Type = &Const{}
+	var _ Type = &Restrict{}
+	var _ Type = &Func{}
+	var _ Type = &FuncProto{}
+	var _ Type = &Var{}
+	var _ Type = &Datasec{}
 }
 
 func TestType(t *testing.T) {
 	types := []func() Type{
 		func() Type { return &Void{} },
-		func() Type { return &Int{} },
+		func() Type { return &Int{Size: 2, Bits: 3} },
 		func() Type { return &Pointer{Target: &Void{}} },
 		func() Type { return &Array{Type: &Int{}} },
 		func() Type {
@@ -118,21 +131,99 @@ func TestType(t *testing.T) {
 	for _, fn := range types {
 		typ := fn()
 		t.Run(fmt.Sprintf("%T", typ), func(t *testing.T) {
+			t.Logf("%v", typ)
+
 			if typ == typ.copy() {
 				t.Error("Copy doesn't copy")
 			}
 
-			var first, second copyStack
+			var first, second typeDeque
 			typ.walk(&first)
 			typ.walk(&second)
 
-			if diff := cmp.Diff(first, second, compareTypes); diff != "" {
+			if diff := cmp.Diff(first.all(), second.all(), compareTypes); diff != "" {
 				t.Errorf("Walk mismatch (-want +got):\n%s", diff)
 			}
 		})
 	}
 }
 
+func TestTypeDeque(t *testing.T) {
+	a, b := new(Type), new(Type)
+
+	t.Run("pop", func(t *testing.T) {
+		var td typeDeque
+		td.push(a)
+		td.push(b)
+
+		if td.pop() != b {
+			t.Error("Didn't pop b first")
+		}
+
+		if td.pop() != a {
+			t.Error("Didn't pop a second")
+		}
+
+		if td.pop() != nil {
+			t.Error("Didn't pop nil")
+		}
+	})
+
+	t.Run("shift", func(t *testing.T) {
+		var td typeDeque
+		td.push(a)
+		td.push(b)
+
+		if td.shift() != a {
+			t.Error("Didn't shift a second")
+		}
+
+		if td.shift() != b {
+			t.Error("Didn't shift b first")
+		}
+
+		if td.shift() != nil {
+			t.Error("Didn't shift nil")
+		}
+	})
+
+	t.Run("push", func(t *testing.T) {
+		var td typeDeque
+		td.push(a)
+		td.push(b)
+		td.shift()
+
+		ts := make([]Type, 12)
+		for i := range ts {
+			td.push(&ts[i])
+		}
+
+		if td.shift() != b {
+			t.Error("Didn't shift b first")
+		}
+		for i := range ts {
+			if td.shift() != &ts[i] {
+				t.Fatal("Shifted wrong Type at pos", i)
+			}
+		}
+	})
+
+	t.Run("all", func(t *testing.T) {
+		var td typeDeque
+		td.push(a)
+		td.push(b)
+
+		all := td.all()
+		if len(all) != 2 {
+			t.Fatal("Expected 2 elements, got", len(all))
+		}
+
+		if all[0] != a || all[1] != b {
+			t.Fatal("Elements don't match")
+		}
+	})
+}
+
 func newCyclicalType(n int) Type {
 	ptr := &Pointer{}
 	prev := Type(ptr)
