@@ -1,38 +1,18 @@
 package testutils
 
 import (
-	"bytes"
 	"errors"
-	"sync"
 	"testing"
 
 	"github.com/cilium/ebpf/internal"
-	"github.com/cilium/ebpf/internal/unix"
 )
 
-var (
-	kernelVersionOnce sync.Once
-	kernelVersion     internal.Version
-)
-
-func mustKernelVersion() internal.Version {
-	kernelVersionOnce.Do(func() {
-		var uname unix.Utsname
-		err := unix.Uname(&uname)
-		if err != nil {
-			panic(err)
-		}
-
-		end := bytes.IndexByte(uname.Release[:], 0)
-		release := string(uname.Release[:end])
-
-		kernelVersion, err = internal.NewVersion(release)
-		if err != nil {
-			panic(err)
-		}
-	})
-
-	return kernelVersion
+func MustKernelVersion() internal.Version {
+	v, err := internal.KernelVersion()
+	if err != nil {
+		panic(err)
+	}
+	return v
 }
 
 func CheckFeatureTest(t *testing.T, fn func() error) {
@@ -67,7 +47,7 @@ func checkKernelVersion(tb testing.TB, ufe *internal.UnsupportedFeatureError) {
 		return
 	}
 
-	kernelVersion := mustKernelVersion()
+	kernelVersion := MustKernelVersion()
 	if ufe.MinimumVersion.Less(kernelVersion) {
 		tb.Helper()
 		tb.Fatalf("Feature '%s' isn't supported even though kernel %s is newer than %s",
@@ -83,7 +63,7 @@ func SkipOnOldKernel(tb testing.TB, minVersion, feature string) {
 		tb.Fatalf("Invalid version %s: %s", minVersion, err)
 	}
 
-	if mustKernelVersion().Less(minv) {
+	if MustKernelVersion().Less(minv) {
 		tb.Skipf("Test requires at least kernel %s (due to missing %s)", minv, feature)
 	}
 }
