@@ -3,7 +3,6 @@ package ebpf
 import (
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"math"
 	"os"
 	"path/filepath"
@@ -12,6 +11,7 @@ import (
 	"testing"
 	"unsafe"
 
+	"github.com/cilium/ebpf/asm"
 	"github.com/cilium/ebpf/internal"
 	"github.com/cilium/ebpf/internal/btf"
 	"github.com/cilium/ebpf/internal/testutils"
@@ -343,7 +343,7 @@ func TestNestedMapPin(t *testing.T) {
 	}
 	defer m.Close()
 
-	tmp, err := ioutil.TempDir("/sys/fs/bpf", "ebpf-test")
+	tmp, err := os.MkdirTemp("/sys/fs/bpf", "ebpf-test")
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -1033,6 +1033,99 @@ func TestPerCPUMarshaling(t *testing.T) {
 	}
 }
 
+type bpfCgroupStorageKey struct {
+	CgroupInodeId uint64
+	AttachType    AttachType
+	_             [4]byte // Padding
+}
+
+func TestCgroupPerCPUStorageMarshaling(t *testing.T) {
+	numCPU, err := internal.PossibleCPUs()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if numCPU < 2 {
+		t.Skip("Test requires at least two CPUs")
+	}
+	testutils.SkipOnOldKernel(t, "5.9", "per-CPU CGoup storage with write from user space support")
+
+	cgroup := testutils.CreateCgroup(t)
+
+	arr, err := NewMap(&MapSpec{
+		Type:      PerCPUCGroupStorage,
+		KeySize:   uint32(unsafe.Sizeof(bpfCgroupStorageKey{})),
+		ValueSize: uint32(unsafe.Sizeof(uint64(0))),
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	t.Cleanup(func() {
+		arr.Close()
+	})
+
+	prog, err := NewProgram(&ProgramSpec{
+		Type:       CGroupSKB,
+		AttachType: AttachCGroupInetEgress,
+		License:    "MIT",
+		Instructions: asm.Instructions{
+			asm.LoadMapPtr(asm.R1, arr.FD()),
+			asm.Mov.Imm(asm.R2, 0),
+			asm.FnGetLocalStorage.Call(),
+			asm.Mov.Imm(asm.R0, 0),
+			asm.Return(),
+		},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer prog.Close()
+
+	progAttachAttrs := internal.BPFProgAttachAttr{
+		TargetFd:     uint32(cgroup.Fd()),
+		AttachBpfFd:  uint32(prog.FD()),
+		AttachType:   uint32(AttachCGroupInetEgress),
+		AttachFlags:  0,
+		ReplaceBpfFd: 0,
+	}
+	err = internal.BPFProgAttach(&progAttachAttrs)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer func() {
+		attr := internal.BPFProgDetachAttr{
+			TargetFd:    uint32(cgroup.Fd()),
+			AttachBpfFd: uint32(prog.FD()),
+			AttachType:  uint32(AttachCGroupInetEgress),
+		}
+		if err := internal.BPFProgDetach(&attr); err != nil {
+			t.Fatal(err)
+		}
+	}()
+
+	var mapKey = &bpfCgroupStorageKey{
+		CgroupInodeId: testutils.GetCgroupIno(t, cgroup),
+		AttachType:    AttachCGroupInetEgress,
+	}
+
+	values := []uint64{1, 2}
+	if err := arr.Put(mapKey, values); err != nil {
+		t.Fatalf("Can't set cgroup %s storage: %s", cgroup.Name(), err)
+	}
+
+	var retrieved []uint64
+	if err := arr.Lookup(mapKey, &retrieved); err != nil {
+		t.Fatalf("Can't retrieve cgroup %s storage: %s", cgroup.Name(), err)
+	}
+
+	for i, want := range []uint64{1, 2} {
+		if retrieved[i] == 0 {
+			t.Errorf("Item %d is 0", i)
+		} else if have := retrieved[i]; have != want {
+			t.Errorf("PerCPUCGroupStorage map is not correctly unmarshaled, expected %d but got %d", want, have)
+		}
+	}
+}
+
 func TestMapMarshalUnsafe(t *testing.T) {
 	m, err := NewMap(&MapSpec{
 		Type:       Hash,
@@ -1404,6 +1497,43 @@ func BenchmarkMarshalling(b *testing.B) {
 	})
 }
 
+func BenchmarkPerCPUMarshalling(b *testing.B) {
+	newMap := func(valueSize uint32) *Map {
+		m, err := NewMap(&MapSpec{
+			Type:       PerCPUHash,
+			KeySize:    8,
+			ValueSize:  valueSize,
+			MaxEntries: 1,
+		})
+		if err != nil {
+			b.Fatal(err)
+		}
+		return m
+	}
+
+	key := uint64(1)
+	val := []uint64{1, 2, 3, 4, 5, 6, 7, 8}
+
+	m := newMap(8)
+	if err := m.Put(key, val[0:]); err != nil {
+		b.Fatal(err)
+	}
+
+	b.Run("reflection", func(b *testing.B) {
+		b.ReportAllocs()
+		b.ResetTimer()
+
+		var value []uint64
+
+		for i := 0; i < b.N; i++ {
+			err := m.Lookup(unsafe.Pointer(&key), &value)
+			if err != nil {
+				b.Fatal("Can't get key:", err)
+			}
+		}
+	})
+}
+
 func BenchmarkMap(b *testing.B) {
 	m, err := NewMap(&MapSpec{
 		Type:       Hash,
