@@ -6,6 +6,7 @@ import (
 	"errors"
 	"fmt"
 	"math"
+	"os"
 	"path/filepath"
 	"runtime"
 	"strings"
@@ -13,6 +14,8 @@ import (
 	"testing"
 	"time"
 
+	qt "github.com/frankban/quicktest"
+
 	"github.com/cilium/ebpf/asm"
 	"github.com/cilium/ebpf/internal"
 	"github.com/cilium/ebpf/internal/testutils"
@@ -180,25 +183,82 @@ func TestProgramClose(t *testing.T) {
 
 func TestProgramPin(t *testing.T) {
 	prog := createSocketFilter(t)
+	c := qt.New(t)
 	defer prog.Close()
 
-	tmp := tempBPFFS(t)
+	tmp := testutils.TempBPFFS(t)
 
 	path := filepath.Join(tmp, "program")
 	if err := prog.Pin(path); err != nil {
 		t.Fatal(err)
 	}
+
+	pinned := prog.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+
 	prog.Close()
 
-	prog, err := LoadPinnedProgram(path)
+	prog, err := LoadPinnedProgram(path, nil)
 	testutils.SkipIfNotSupported(t, err)
 	if err != nil {
 		t.Fatal(err)
 	}
 	defer prog.Close()
 
-	if prog.abi.Type != SocketFilter {
-		t.Error("Expected pinned program to have type SocketFilter, but got", prog.abi.Type)
+	if prog.Type() != SocketFilter {
+		t.Error("Expected pinned program to have type SocketFilter, but got", prog.Type())
+	}
+
+	if !prog.IsPinned() {
+		t.Error("Expected IsPinned to be true")
+	}
+}
+
+func TestProgramUnpin(t *testing.T) {
+	prog := createSocketFilter(t)
+	c := qt.New(t)
+	defer prog.Close()
+
+	tmp := testutils.TempBPFFS(t)
+
+	path := filepath.Join(tmp, "program")
+	if err := prog.Pin(path); err != nil {
+		t.Fatal(err)
+	}
+
+	pinned := prog.IsPinned()
+	c.Assert(pinned, qt.Equals, true)
+
+	if err := prog.Unpin(); err != nil {
+		t.Fatal("Failed to unpin program:", err)
+	}
+	if _, err := os.Stat(path); err == nil {
+		t.Fatal("Pinned program path still exists after unpinning:", err)
+	}
+}
+
+func TestProgramLoadPinnedWithFlags(t *testing.T) {
+	// Introduced in commit 6e71b04a8224.
+	testutils.SkipOnOldKernel(t, "4.14", "file_flags in BPF_OBJ_GET")
+
+	prog := createSocketFilter(t)
+	defer prog.Close()
+
+	tmp := testutils.TempBPFFS(t)
+
+	path := filepath.Join(tmp, "program")
+	if err := prog.Pin(path); err != nil {
+		t.Fatal(err)
+	}
+
+	prog.Close()
+
+	_, err := LoadPinnedProgram(path, &LoadPinOptions{
+		Flags: math.MaxUint32,
+	})
+	testutils.SkipIfNotSupported(t, err)
+	if !errors.Is(err, unix.EINVAL) {
+		t.Fatal("Invalid flags don't trigger an error:", err)
 	}
 }
 
@@ -214,7 +274,7 @@ func TestProgramVerifierOutputOnError(t *testing.T) {
 		t.Fatal("Expected program to be invalid")
 	}
 
-	if strings.Index(err.Error(), "exit") == -1 {
+	if !strings.Contains(err.Error(), "exit") {
 		t.Error("No verifier output in error message")
 	}
 }
@@ -270,6 +330,23 @@ func TestProgramVerifierOutput(t *testing.T) {
 	}
 }
 
+func TestProgramWithUnsatisfiedReference(t *testing.T) {
+	coll, err := LoadCollectionSpec("testdata/loader-el.elf")
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// The program will have at least one map reference.
+	progSpec := coll.Programs["xdp_prog"]
+	progSpec.ByteOrder = nil
+
+	_, err = NewProgram(progSpec)
+	if !errors.Is(err, errUnsatisfiedReference) {
+		t.Fatal("Expected an error wrapping errUnsatisfiedReference, got", err)
+	}
+	t.Log(err)
+}
+
 func TestProgramName(t *testing.T) {
 	if err := haveObjName(); err != nil {
 		t.Skip(err)
@@ -278,7 +355,7 @@ func TestProgramName(t *testing.T) {
 	prog := createSocketFilter(t)
 	defer prog.Close()
 
-	info, err := bpfGetProgInfoByFD(prog.fd)
+	info, err := bpfGetProgInfoByFD(prog.fd, nil)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -379,7 +456,12 @@ func TestProgramFromFD(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	prog2.Close()
+	// Both programs refer to the same fd now. Closing either of them will
+	// release the fd to the OS, which then might re-use that fd for another
+	// test. Once we close the second map we might close the re-used fd
+	// inadvertently, leading to spurious test failures.
+	// To avoid this we have to "leak" one of the programs.
+	prog2.fd.Forget()
 }
 
 func TestHaveProgTestRun(t *testing.T) {
@@ -473,22 +555,133 @@ func TestProgramRejectIncorrectByteOrder(t *testing.T) {
 	}
 }
 
-func TestProgramTypeLSM(t *testing.T) {
-	prog, err := NewProgram(&ProgramSpec{
-		AttachTo:   "task_getpgid",
-		AttachType: AttachLSMMac,
+func TestProgramSpecTag(t *testing.T) {
+	arr := createArray(t)
+	defer arr.Close()
+
+	spec := &ProgramSpec{
+		Type: SocketFilter,
 		Instructions: asm.Instructions{
-			asm.LoadImm(asm.R0, 0, asm.DWord),
+			asm.LoadImm(asm.R0, -1, asm.DWord),
+			asm.LoadMapPtr(asm.R1, arr.FD()),
+			asm.Mov.Imm32(asm.R0, 0),
 			asm.Return(),
 		},
-		License: "GPL",
-		Type:    LSM,
-	})
+		License: "MIT",
+	}
+
+	prog, err := NewProgram(spec)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer prog.Close()
+
+	info, err := prog.Info()
 	testutils.SkipIfNotSupported(t, err)
 	if err != nil {
 		t.Fatal(err)
 	}
-	prog.Close()
+
+	tag, err := spec.Tag()
+	if err != nil {
+		t.Fatal("Can't calculate tag:", err)
+	}
+
+	if tag != info.Tag {
+		t.Errorf("Calculated tag %s doesn't match kernel tag %s", tag, info.Tag)
+	}
+}
+
+func TestProgramTypeLSM(t *testing.T) {
+	lsmTests := []struct {
+		attachFn    string
+		flags       uint32
+		expectedErr bool
+	}{
+		{
+			attachFn: "task_getpgid",
+		},
+		{
+			attachFn:    "task_setnice",
+			flags:       unix.BPF_F_SLEEPABLE,
+			expectedErr: true,
+		},
+		{
+			attachFn: "file_open",
+			flags:    unix.BPF_F_SLEEPABLE,
+		},
+	}
+	for _, tt := range lsmTests {
+		t.Run(tt.attachFn, func(t *testing.T) {
+			prog, err := NewProgram(&ProgramSpec{
+				AttachTo:   tt.attachFn,
+				AttachType: AttachLSMMac,
+				Instructions: asm.Instructions{
+					asm.LoadImm(asm.R0, 0, asm.DWord),
+					asm.Return(),
+				},
+				License: "GPL",
+				Type:    LSM,
+				Flags:   tt.flags,
+			})
+			testutils.SkipIfNotSupported(t, err)
+
+			if tt.flags&unix.BPF_F_SLEEPABLE != 0 {
+				testutils.SkipOnOldKernel(t, "5.11", "BPF_F_SLEEPABLE for LSM progs")
+			}
+			if tt.expectedErr && err == nil {
+				t.Errorf("Test case '%s': expected error", tt.attachFn)
+			}
+			if !tt.expectedErr && err != nil {
+				t.Errorf("Test case '%s': expected success", tt.attachFn)
+			}
+			prog.Close()
+		})
+	}
+}
+
+func TestProgramTargetBTF(t *testing.T) {
+	// Load a file that contains valid BTF, but doesn't contain the types
+	// we need for bpf_iter.
+	fh, err := os.Open("testdata/invalid_btf_map_init-el.elf")
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer fh.Close()
+
+	reader := &testReaderAt{file: fh}
+
+	prog, err := NewProgramWithOptions(&ProgramSpec{
+		Type:       Tracing,
+		AttachType: AttachTraceIter,
+		AttachTo:   "bpf_map",
+		Instructions: asm.Instructions{
+			asm.Mov.Imm(asm.R0, 0),
+			asm.Return(),
+		},
+		License: "MIT",
+	}, ProgramOptions{
+		TargetBTF: reader,
+	})
+	if err == nil {
+		prog.Close()
+	}
+	if !errors.Is(err, ErrNotSupported) {
+		t.Error("Expected ErrNotSupported, got", err)
+	}
+	if !reader.read {
+		t.Error("TargetBTF is not read")
+	}
+}
+
+type testReaderAt struct {
+	file *os.File
+	read bool
+}
+
+func (ra *testReaderAt) ReadAt(p []byte, off int64) (int, error) {
+	ra.read = true
+	return ra.file.ReadAt(p, off)
 }
 
 func createProgramArray(t *testing.T) *Map {
@@ -555,7 +748,7 @@ func ExampleNewProgramWithOptions() {
 
 // It's possible to read a program directly from a ProgramArray.
 func ExampleProgram_unmarshalFromMap() {
-	progArray, err := LoadPinnedMap("/path/to/map")
+	progArray, err := LoadPinnedMap("/path/to/map", nil)
 	if err != nil {
 		panic(err)
 	}
@@ -584,3 +777,24 @@ func ExampleProgram_unmarshalFromMap() {
 		panic(err)
 	}
 }
+
+func ExampleProgramSpec_Tag() {
+	spec := &ProgramSpec{
+		Type: SocketFilter,
+		Instructions: asm.Instructions{
+			asm.LoadImm(asm.R0, 0, asm.DWord),
+			asm.Return(),
+		},
+		License: "MIT",
+	}
+
+	prog, _ := NewProgram(spec)
+	info, _ := prog.Info()
+	tag, _ := spec.Tag()
+
+	if info.Tag != tag {
+		fmt.Printf("The tags don't match: %s != %s\n", info.Tag, tag)
+	} else {
+		fmt.Println("The programs are identical, tag is", tag)
+	}
+}
